#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>

#include "plugin.hpp"
#include "xCreate.hpp"
#include "types.hpp"
#include "tools.hpp"
#include "tools_xcreate.hpp"
#include "creator.hpp"
#include "lang.hpp"

PluginStartupInfo startupInfo;
static FarStandardFunctions FSF;

static HANDLE Settings;

static wchar_t   BOOTFile[256];
static wchar_t   comment[256];
static int    isDS;
static int    imageType;
static int    isWriteBOOT;
static int    format;
static int    interleave;
static int    writeProtect;
static int    isComment;

static wchar_t   title[12];
static wchar_t   fileName[256];

extern const GUID MainGuid = __uuidof(struct __declspec(uuid("{0CB259A1-7CB1-4B6E-AA68-6AE80AC93CA9}")) TMainGuid);

void WINAPI _export SetStartupInfoW(const struct PluginStartupInfo *info)
{
    if(info->StructSize < sizeof(PluginStartupInfo))
    {
        return;
    }
    //    PluginStartupInfo
    startupInfo = *info;
    FSF = *(info->FSF);
    startupInfo.FSF = &FSF;

    FarSettingsCreate fsc{ sizeof(FarSettingsCreate), MainGuid, 0 };
    startupInfo.SettingsControl(INVALID_HANDLE_VALUE, SCTL_CREATE, PSL_LOCAL, &fsc);
    Settings = fsc.Handle;

    FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

    fsi.Type = FST_QWORD;

    fsi.Name = L"IsDS";
    fsi.Number = 1;
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    isDS = fsi.Number;

    fsi.Name = L"IsWriteBOOT";
    fsi.Number = 0;
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    isWriteBOOT = fsi.Number;

    fsi.Name = L"DefaultFormat";
    fsi.Number = FMTI_TRD;
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    format = fsi.Number;

    fsi.Name = L"Interleave";
    fsi.Number = 2;
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    interleave = fsi.Number;

    fsi.Name = L"IsComment";
    fsi.Number = 1;
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    isComment = fsi.Number;

    fsi.Type = FST_STRING;
    fsi.Name = L"DefaultBOOT";
    fsi.String = L"";
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    wcsncpy(BOOTFile, fsi.String, std::size(BOOTFile));
    BOOTFile[std::size(BOOTFile) - 1] = 0;

    fsi.Name = L"DefaultComment";
    fsi.String = GetMsg(MDefComment);
    startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
    wcsncpy(comment, fsi.String, std::size(comment));
    comment[std::size(comment) - 1] = 0;
}

void WINAPI _export ExitFARW()
{
    startupInfo.SettingsControl(Settings, SCTL_FREE, 0, nullptr);
    Settings = nullptr;
}


HANDLE WINAPI _export OpenW(const struct OpenInfo *Info)
{
  FarDialogItem dialogItems[] =
  {
      /*0*/ {DI_DOUBLEBOX,     3,  1, 63, 23, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MCreate) },
      /*1*/ {DI_TEXT,          5,  2,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MFileName) },
      /*2*/ {DI_EDIT,          5,  3, 61,  0,  { TRUE }, L"xCreateFileName", nullptr,                         DIF_HISTORY, L"" },

      /*3*/ {DI_TEXT,          3,  4,  0,  0, { FALSE }, nullptr,            nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
      /*4*/ {DI_TEXT,          5,  5,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MFileFormat) },
      /*5*/ {DI_RADIOBUTTON,   5,  6,  0,  0, { FALSE }, nullptr,            nullptr,                           DIF_GROUP, GetMsg(MTRD) },
      /*6*/ {DI_RADIOBUTTON,  14,  6,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MFDI) },
      /*7*/ {DI_RADIOBUTTON,  23,  6,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MSCL) },
      /*8*/ {DI_RADIOBUTTON,  32,  6,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MFDD) },
      /*9*/ {DI_RADIOBUTTON,  41,  6,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MUDI) },
     /*10*/ {DI_RADIOBUTTON,  50,  6,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MTeleDisk) },

     /*11*/ {DI_TEXT,          3,  7,  0,  0, { FALSE }, nullptr,            nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L""},
     /*12*/ {DI_TEXT,          5,  8,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MTitle) },
     /*13*/ {DI_EDIT,          5,  9, 16,  0, { FALSE }, L"xCreateTitle",    nullptr,                         DIF_HISTORY, L"" },

     /*14*/ {DI_TEXT,          3, 10,  0,  0, { FALSE }, nullptr,            nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
     /*15*/ {DI_TEXT,          5, 11,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MInterleave) },
     /*16*/ {DI_RADIOBUTTON,   5, 12,  0,  0, { FALSE }, nullptr,            nullptr,                           DIF_GROUP, GetMsg(MInterleave1) },
     /*17*/ {DI_RADIOBUTTON,   5, 13,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MInterleave2) },

     /*18*/ {DI_TEXT,          3, 14,  0,  0, { FALSE }, nullptr,            nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
     /*19*/ {DI_CHECKBOX,      5, 15,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MWriteProtect) },
     /*20*/ {DI_CHECKBOX,      5, 16,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MInstallDS) },
     /*21*/ {DI_CHECKBOX,      5, 17,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MWriteBOOT) },
     /*22*/ {DI_EDIT,          5, 18, 61,  0, { FALSE }, L"xCreateBOOTName", nullptr,                         DIF_HISTORY, BOOTFile },
     /*23*/ {DI_CHECKBOX,      5, 19,  0,  0, { FALSE }, nullptr,            nullptr,                                   0, GetMsg(MComment) },
     /*24*/ {DI_EDIT,          5, 20, 61,  0, { FALSE }, L"xCreateComment",  nullptr,                         DIF_HISTORY, comment },

     /*25*/ {DI_TEXT,          3, 21,  0,  0, { FALSE }, nullptr,            nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
     /*26*/ {DI_BUTTON,        0, 22,  0,  0, { FALSE }, nullptr,            nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk) },
     /*27*/ {DI_BUTTON,        0, 22,  0,  0, { FALSE }, nullptr,            nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };

  dialogItems[19].Selected = writeProtect;
  dialogItems[20].Selected = isDS;
  dialogItems[21].Selected = isWriteBOOT;
  dialogItems[23].Selected = isComment;
  if(interleave == 1)
    dialogItems[16].Selected = TRUE;
  else
    dialogItems[17].Selected = TRUE;

  switch(format)
  {
    case FMTI_TRD:
              dialogItems[5].Selected = TRUE;
              break;
    case FMTI_FDI:
              dialogItems[6].Selected = TRUE;
              break;
    case FMTI_SCL:
              dialogItems[7].Selected = TRUE;
              break;
    case FMTI_FDD:
              dialogItems[8].Selected = TRUE;
              break;
    case FMTI_UDI:
              dialogItems[9].Selected = TRUE;
              break;
    case FMTI_TD:
              dialogItems[10].Selected = TRUE;
              break;
    default:
              dialogItems[5].Selected = TRUE;
  }

  std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                   &__uuidof(struct __declspec(uuid("{2393AFA0-EA07-43AE-8EB3-3FF566A93985}")) TDlgGuid),
                                   -1, -1, 67, 25,
                                   nullptr,
                                   dialogItems,
                                   std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
  if(Dlg.get() == INVALID_HANDLE_VALUE)
  {
      return nullptr;
  }
  intptr_t askCode = startupInfo.DialogRun(Dlg.get());

  if(askCode != 26) return nullptr;

  writeProtect = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 19, 0);
  isDS         = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 20, 0);
  isWriteBOOT  = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 21, 0);
  isComment    = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 23, 0);

  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 16, 0)) interleave = 1;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 17, 0)) interleave = 2;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 5, 0)) format = FMTI_TRD;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 6, 0)) format = FMTI_FDI;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0)) format = FMTI_SCL;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 8, 0)) format = FMTI_FDD;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 9, 0)) format = FMTI_UDI;
  if(startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 10, 0)) format = FMTI_TD;

  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(comment) - 1, comment };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 24, &ItemData);
  }

  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(BOOTFile) - 1, BOOTFile };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 22, &ItemData);
  }

  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(title) - 1, title };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 13, &ItemData);
  }

  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(fileName) - 1, fileName };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
  }

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

  fsi.Type = FST_STRING;
  fsi.Name = L"DefaultComment";
  fsi.String = comment;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"DefaultBOOT";
  fsi.String = BOOTFile;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Type = FST_QWORD;
  fsi.Name = L"IsDS";
  fsi.Number = isDS;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"IsWriteBOOT";
  fsi.Number = isWriteBOOT;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"DefaultFormat";
  fsi.Number = format;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"Interleave";
  fsi.Number = interleave;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"WriteProtect";
  fsi.Number = writeProtect;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"IsComment";
  fsi.Number = isComment;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  std::wstring FileNameFull;
  {
      size_t Size = FSF.ConvertPath(CPM_NATIVE, fileName, nullptr, 0);
      FileNameFull.resize(Size);
      FSF.ConvertPath(CPM_NATIVE, fileName, FileNameFull.data(), FileNameFull.size());
  }

  std::wstring BootFileFull;
  if(isWriteBOOT)
  {
      size_t Size = FSF.ConvertPath(CPM_NATIVE, BOOTFile, nullptr, 0);
      BootFileFull.resize(Size);
      FSF.ConvertPath(CPM_NATIVE, BOOTFile, BootFileFull.data(), BootFileFull.size());
  }

  create(FileNameFull.c_str(), title, writeProtect, isDS, format, interleave, isWriteBOOT?BootFileFull.c_str():0, isComment&&comment[0]?comment:0);

  startupInfo.PanelControl(PANEL_ACTIVE, FCTL_UPDATEPANEL, 1, nullptr);
  startupInfo.PanelControl(PANEL_ACTIVE, FCTL_REDRAWPANEL, 0, nullptr);
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_REDRAWPANEL, 0, nullptr);

  return nullptr;
}

struct __declspec(uuid("{F9931BA1-4A53-4B04-B6A3-9971B827C41B}")) TMenuGuid {};
static const GUID MenuGuid = __uuidof(TMenuGuid);

void WINAPI _export GetPluginInfoW(PluginInfo *info)
{
    info->StructSize = sizeof(*info);
    info->Flags = 0;

    static const wchar_t *menuStrings[1];
    menuStrings[0] = GetMsg(MCreate);

    info->PluginMenu.Guids = &MenuGuid;
    info->PluginMenu.Strings = menuStrings;
    info->PluginMenu.Count = std::size(menuStrings);
}

void WINAPI _export GetGlobalInfoW(struct GlobalInfo *Info)
{
    Info->StructSize = sizeof(struct GlobalInfo);
    Info->MinFarVersion = MAKEFARVERSION(3, 0, 0, 2927, VS_RELEASE);   //   Far Manager,    API
    Info->Version = MAKEFARVERSION(1, 0, 0, 1, VS_RC);   //   1.0.0.1, -
    Info->Guid = MainGuid;
    Info->Title = L"xCreate";
    Info->Description = L"xCreate";
    Info->Author = L"Elf/2 & Alexandr Medvedev";
}
