#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <iterator>

#include "types.hpp"
#include "types_isd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_isd.hpp"
#include "iSDOS.hpp"
#include "iSDOS_tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern const GUID MainGuid;

constexpr auto PKF_CONTROL{ LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED };

int Manager::processKey(const INPUT_RECORD *Rec)
{
    if(Rec->EventType != KEY_EVENT)
    {
        return FALSE;
    }

    if(!Rec->Event.KeyEvent.bKeyDown)
    {
        return FALSE;
    }
    auto key = Rec->Event.KeyEvent.wVirtualKeyCode;
    auto controlState = Rec->Event.KeyEvent.dwControlKeyState;

  if(key == VK_RETURN && !(controlState & PKF_CONTROL))
  {
    intptr_t Size = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, nullptr);
    std::unique_ptr<PluginPanelItem, decltype(::free) *> Item((PluginPanelItem *)::malloc(Size), ::free);
    FarGetPluginPanelItem GetItem = { sizeof(FarGetPluginPanelItem), (size_t)Size, Item.get() };
    intptr_t r = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, &GetItem);
    if(r == FALSE)
    {
        return FALSE;
    }
    const wchar_t* fName = Item->FileName;

    int fNum = findFile(files, fName);
    if (strnicmp(files[fNum].ext,"com",3) == 0) return TRUE;
    if (strnicmp(files[fNum].ext,"bat",3) == 0) return TRUE;
    if (strnicmp(files[fNum].ext,"exe",3) == 0) return TRUE;
    if (strnicmp(files[fNum].ext,"cmd",3) == 0) return TRUE;
  }

  if((controlState & PKF_CONTROL) && key == 0x41)
  {
    HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);
    PanelInfo info = { sizeof(PanelInfo) };
    startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &info);
    startupInfo.RestoreScreen(screen);
    
    if(info.ItemsNumber > 0 && info.CurrentItem > 0)
    {
      intptr_t Size = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, nullptr);
      std::unique_ptr<PluginPanelItem, decltype(::free)*> Item((PluginPanelItem *)::malloc(Size), ::free);
      FarGetPluginPanelItem GetItem = { sizeof(FarGetPluginPanelItem), (size_t)Size, Item.get() };
      startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, &GetItem);

      const wchar_t *fName = Item->FileName;
      int fNum = findFile(files, fName);
      if(!fNum) return TRUE;

      wchar_t name[8+3+2] = L"";
      makeName(files[fNum], name);

      int days[12]={31,28,31,30,31,30,31,31,30,31,30,31};
      wchar_t day[3], month[3], year[5];
      SYSTEMTIME st = makeDate(files[fNum]);
      wsprintfW(day,  L"%02d",st.wDay);
      wsprintfW(month,L"%02d",st.wMonth);
      wsprintfW(year, L"%04d",st.wYear);

      if (isDir(files[fNum]))
      {
        FarDialogItem dialogItems[]=
        {
          { DI_DOUBLEBOX, 3,  1, 41, 14, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MAttributes) },
          { DI_TEXT,      5,  2,  0,  0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MAttrsFor) },
          { DI_TEXT,      5,  3,  0,  0,  { TRUE }, nullptr, nullptr,                     DIF_CENTERGROUP, name },
          { DI_TEXT,      3,  4,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_CHECKBOX,  5,  5,  0,  0,  { TRUE }, nullptr, nullptr,                                   0, GetMsg(MReadProtect) },
          { DI_CHECKBOX,  5,  6,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MWriteProtect) },
          { DI_CHECKBOX,  5,  7,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDeleteProtect) },
          { DI_CHECKBOX,  5,  8,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MHidden) },
          { DI_TEXT,      3,  9,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_TEXT,     10, 10,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDDMMYYYY) },
          { DI_TEXT,      5, 11,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDate) },
          { DI_FIXEDIT,  10, 11, 11, 13, { FALSE }, nullptr, nullptr,                                   0, day },
          { DI_TEXT,     12, 11,  0,  0, { FALSE }, nullptr, nullptr,                                   0, L"." },
          { DI_FIXEDIT,  13, 11, 14, 13, { FALSE }, nullptr, nullptr,                                   0, month },
          { DI_TEXT,     15, 11,  0,  0, { FALSE }, nullptr, nullptr,                                   0, L"." },
          { DI_FIXEDIT,  16, 11, 19, 13, { FALSE }, nullptr, nullptr,                                   0, year },

          { DI_TEXT,     3,  12,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_BUTTON,   0,  13,  0,  0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk) },
          { DI_BUTTON,   0,  13,  0,  0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
        };

        dialogItems[4].Selected = files[fNum].attr & FLAG_READ_PROTECT;
        dialogItems[5].Selected = files[fNum].attr & FLAG_WRITE_PROTECT;
        dialogItems[6].Selected = files[fNum].attr & FLAG_DELETE_PROTECT;
        dialogItems[7].Selected = files[fNum].attr & FLAG_HIDDEN;
   
        std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                         &__uuidof(struct __declspec(uuid("{434BB78B-4FE6-47F6-9593-A2FB72475014}")) TDlgGuid),
                                         -1, -1, 45, 16,
                                         NULL,
                                         dialogItems,
                                         std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
        if(Dlg.get() == INVALID_HANDLE_VALUE)
        {
            return FALSE;
        }
        intptr_t askCode = startupInfo.DialogRun(Dlg.get());

        if(askCode != 17) return TRUE;
   
        files[fNum].attr &= 0xFF & ~(FLAG_READ_PROTECT | FLAG_WRITE_PROTECT | FLAG_DELETE_PROTECT | FLAG_HIDDEN);
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0)) files[fNum].attr |= FLAG_READ_PROTECT;
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 5, 0)) files[fNum].attr |= FLAG_WRITE_PROTECT;
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 6, 0)) files[fNum].attr |= FLAG_DELETE_PROTECT;
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0)) files[fNum].attr |= FLAG_HIDDEN;

        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(day) - 1, day };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 11, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(month) - 1, month };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 13, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(year) - 1, year };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 15, &ItemData);
        }

        int result = str2int(month, 2);
        if((result <= 12)&&(result > 0)) st.wMonth = result;
        result = str2int(year, 4);
        if(result != -1)
        {
          if(result < 80) st.wYear = result + 2000;
          if((result >= 80)&&(result < 100)) st.wYear = result + 1900;
          if((result >= 1980)&&(result < 2100)) st.wYear = result;
        }
        result = str2int(day, 2);
        if(result >0)
        {
          if(result <= days[st.wMonth-1]) st.wDay = result;
          if((result == 29)&&(st.wMonth==2)&&(!(st.wYear%4))) st.wDay = result;
        }
      }
      else
      {
        wchar_t stAdr[6], hour[3], minute[3], second[3];
        wsprintfW(hour, L"%02d", st.wHour);
        wsprintfW(minute, L"%02d", st.wMinute);
        wsprintfW(second, L"%02d", st.wSecond);
        wsprintfW(stAdr, L"%d", files[fNum].file.loadAddr);

        FarDialogItem dialogItems[]=
        {
          { DI_DOUBLEBOX, 3,  1, 41, 16, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MAttributes) },
          { DI_TEXT,      5,  2,  0,  0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MAttrsFor) },
          { DI_TEXT,      5,  3,  0,  0,  { TRUE }, nullptr, nullptr,                     DIF_CENTERGROUP, name },
          { DI_TEXT,      3,  4,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_CHECKBOX,  5,  5,  0,  0,  { TRUE }, nullptr, nullptr,                                   0, GetMsg(MReadProtect) },
          { DI_CHECKBOX,  5,  6,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MWriteProtect) },
          { DI_CHECKBOX,  5,  7,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDeleteProtect) },
          { DI_CHECKBOX,  5,  8,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MHidden) },
          { DI_TEXT,      3,  9,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_TEXT,      5, 10,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MStartAdr) },
          { DI_EDIT,     34, 10, 39, 10, { FALSE }, nullptr, nullptr,                                   0, stAdr },
          { DI_TEXT,      3, 11,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_TEXT,     10, 12,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDDMMYYYY) },
          { DI_TEXT,     32, 12,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(Mhhmmss) },
          { DI_TEXT,      5, 13,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDate) },
          { DI_FIXEDIT,  10, 13, 11, 13, { FALSE }, nullptr, nullptr,                                   0, day },
          { DI_TEXT,     12, 13,  0,  0, { FALSE }, nullptr, nullptr,                                   0, L"." },
          { DI_FIXEDIT,  13, 13, 14, 13, { FALSE }, nullptr, nullptr,                                   0, month },
          { DI_TEXT,     15, 13,  0,  0, { FALSE }, nullptr, nullptr,                                   0, L"." },
          { DI_FIXEDIT,  16, 13, 19, 13, { FALSE }, nullptr, nullptr,                                   0, year },
          { DI_TEXT,     26, 13,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MTime) },
          { DI_FIXEDIT,  32, 13, 33, 13, { FALSE }, nullptr, nullptr,                                   0, hour },
          { DI_TEXT,     34, 13,  0,  0, { FALSE }, nullptr, nullptr,                                   0, L":" },
          { DI_FIXEDIT,  35, 13, 36, 13, { FALSE }, nullptr, nullptr,                                   0, minute },
          { DI_TEXT,     37, 13,  0,  0, { FALSE }, nullptr, nullptr,                                   0, L":" },
          { DI_FIXEDIT,  38, 13, 39, 13, { FALSE }, nullptr, nullptr,                                   0, second },
          { DI_TEXT,      3, 14,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
          { DI_BUTTON,    0, 15,  0,  0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk) },
          { DI_BUTTON,    0, 15,  0,  0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
        };

        dialogItems[4].Selected = files[fNum].attr & FLAG_READ_PROTECT;
        dialogItems[5].Selected = files[fNum].attr & FLAG_WRITE_PROTECT;
        dialogItems[6].Selected = files[fNum].attr & FLAG_DELETE_PROTECT;
        dialogItems[7].Selected = files[fNum].attr & FLAG_HIDDEN;

        std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                         &__uuidof(struct __declspec(uuid("{EFAC9D08-9CC4-484E-A492-9C6A05912F65}")) TDlgGuid),
                                         -1, -1, 45, 18,
                                         NULL,
                                         dialogItems,
                                         std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
        if(Dlg.get() == INVALID_HANDLE_VALUE)
        {
            return FALSE;
        }
        intptr_t askCode = startupInfo.DialogRun(Dlg.get());

        if(askCode != 27) return TRUE;
        if(files[fNum].attr == 0xFF && files[fNum].file.systemFlag == 0xFF) return TRUE;
   
        files[fNum].attr &= 0xFF & ~(FLAG_READ_PROTECT | FLAG_WRITE_PROTECT | FLAG_DELETE_PROTECT | FLAG_HIDDEN);
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0)) files[fNum].attr |= FLAG_READ_PROTECT;
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 5, 0)) files[fNum].attr |= FLAG_WRITE_PROTECT;
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 6, 0)) files[fNum].attr |= FLAG_DELETE_PROTECT;
        if (startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0)) files[fNum].attr |= FLAG_HIDDEN;

        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(stAdr) - 1, stAdr };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 10, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(day) - 1, day };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 15, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(month) - 1, month };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 17, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(year) - 1, year };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 19, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(hour) - 1, hour };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 21, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(minute) - 1, minute };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 23, &ItemData);
        }
        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(second) - 1, second };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 25, &ItemData);
        }

        int result = str2int(stAdr, 5);
        if((result != -1)&&(result < 65536)) files[fNum].file.loadAddr = result;

        result = str2int(month, 2);
        if((result <= 12)&&(result > 0)) st.wMonth = result;
        result = str2int(year, 4);
        if(result != -1)
        {
          if(result < 80) st.wYear = result + 2000;
          if((result >= 80)&&(result < 100)) st.wYear = result + 1900;
          if((result >= 1980)&&(result < 2100)) st.wYear = result;
        }
        result = str2int(day, 2);
        if(result >0)
        {
          if(result <= days[st.wMonth-1]) st.wDay = result;
          if((result == 29)&&(st.wMonth==2)&&(!(st.wYear%4))) st.wDay = result;
        }

        result = str2int(hour, 2);
        if((result < 24)&&(result != -1)) st.wHour = result;
        result = str2int(minute, 2);
        if((result < 60)&&(result != -1)) st.wMinute = result;
        result = str2int(second, 2);
        if((result < 60)&&(result != -1)) st.wSecond = result;
      }

      if(st.wDay > days[st.wMonth-1])
        if ((st.wDay > 29)&&(st.wMonth==2)&&(!(st.wYear%4))) st.wDay = 29;
        else st.wDay = days[st.wMonth-1];

      setDate(files[fNum],st);
      if(!openHostFile()) return TRUE;
      writeFolder(files);
      closeHostFile();
      startupInfo.PanelControl(this, FCTL_UPDATEPANEL, 1, nullptr);
      startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
      return TRUE;
    }
  return FALSE;
  }

  if((controlState & SHIFT_PRESSED) && key == VK_F6)
  {
    HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);
    PanelInfo info = { sizeof(PanelInfo) };
    startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &info);
    startupInfo.RestoreScreen(screen);
    
    if(info.ItemsNumber > 0 && info.CurrentItem > 0)
    {
      intptr_t Size = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, nullptr);
      std::unique_ptr<PluginPanelItem, decltype(::free) *> Item((PluginPanelItem *)::malloc(Size), ::free);
      FarGetPluginPanelItem GetItem = { sizeof(FarGetPluginPanelItem), (size_t)Size, Item.get() };
      startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, &GetItem);

      const wchar_t *fName = Item->FileName;
      int fNum = findFile(files, fName);
      if(!fNum) return TRUE;

      wchar_t name[8+3+2] = L"";
      makeName(files[fNum], name);

      wchar_t buf[100];
      wsprintfW(buf, GetMsg(MRenameTo), name);
      FarDialogItem dialogItems[]=
      {
        { DI_DOUBLEBOX, 3, 1, 50, 6, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MRename) },
        { DI_TEXT,      5, 2,  0, 0, { FALSE }, nullptr, nullptr,                                   0, buf },
        { DI_EDIT,      5, 3, 48, 3,  { TRUE }, nullptr, nullptr,                                   0, name, std::size(name) - 1 },

        { DI_TEXT,      3, 4,  0, 0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
        { DI_BUTTON,    0, 5,  0, 0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk) },
        { DI_BUTTON,    0, 5,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
      };

      std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                       &__uuidof(struct __declspec(uuid("{916907F8-DB0E-4E5F-A7C9-3466F5CDDF9B}")) TDlgGuid),
                                       -1, -1, 54, 8,
                                       NULL,
                                       dialogItems,
                                       std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
      if(Dlg.get() == INVALID_HANDLE_VALUE)
      {
          return FALSE;
      }
      intptr_t askCode = startupInfo.DialogRun(Dlg.get());

      if(askCode != 4) return TRUE;

      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(name) - 1, name };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);

      if(isDir(files[fNum]))
        makeCompatibleFolderName(name, name);
      else
      {
        if(files[fNum].attr == 0xFF && files[fNum].file.systemFlag == 0xFF)
          return TRUE;
        makeCompatibleFileName(name, name);
      }

      const wchar_t *p = name;
      if(*p == 0) return TRUE;
      std::fill(std::begin(files[fNum].name), std::end(files[fNum].name), ' ');
      std::fill(std::begin(files[fNum].ext), std::end(files[fNum].ext), ' ');

      {
          unsigned i = 0;
          for(auto c : name)
          {
              if(c == 0 || c == L'.') break;
              //   cp866
              WideCharToMultiByte(866, 0, &c, 1, &files[fNum].name[i++], 1, nullptr, nullptr);
          }
      }

      p = pointToExt(name);
      
      if(p)
      {
        for(int i = 0; i < 3; ++i)
        {
          if(*p == 0) break;

          //   cp866
          WideCharToMultiByte(866, 0, p++, 1, &files[fNum].ext[i], 1, nullptr, nullptr);
        }
      }
      if(!openHostFile()) return TRUE;
      writeFolder(files);
      closeHostFile();
      startupInfo.PanelControl(this, FCTL_UPDATEPANEL, 1, nullptr);
      startupInfo.PanelControl(this, FCTL_REDRAWPANEL, 0, nullptr);
      startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
      return TRUE;
    }
  }
  return FALSE;
}
