#include <windows.h>
#include <iterator>

#include "types.hpp"
#include "association.hpp"
#include "tools.hpp"

#include "debug.hpp"

Association _associations[256];
int         _no_assosiations = 0;

static DWORD _file_size_wc = 0;
static wchar_t *_data = 0;
static wchar_t *_p = 0;

static bool get_line(wchar_t* buf, int max_size = 100)
{
    if(_p - _data >= _file_size_wc) return false;  // EOF

    wchar_t* to = buf;
    while(true)
    {
        if(_p - _data >= _file_size_wc) break; // EOF
        if(_p[0] == 0x0d && _p[1] == 0x0a) { _p += 2; break; }  // EOL
        *to++ = *_p++;

        if(to - buf == max_size - 1) break; // buffer full
    }

    *to = 0;
    return true;
}

static wchar_t* skip_white_spaces(wchar_t* p)
{
    while(*p && (*p == ' ' || *p == '\t')) ++p;
    return p;
}

bool read_associations(const wchar_t* path)
{
    wchar_t full_path[300];
    lstrcpyW(full_path, path);
    lstrcpyW(pointToName(full_path), L"associat.ion");

    HANDLE file = CreateFileW(full_path, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0);

    // file not found
    if(file == INVALID_HANDLE_VALUE) return false;

    _file_size_wc = SetFilePointer(file, 0, 0, FILE_END) / sizeof(wchar_t);
    SetFilePointer(file, 0, 0, FILE_BEGIN);

    _data = new wchar_t[_file_size_wc + 1];
    DWORD read;
    ReadFile(file, _data, _file_size_wc * sizeof(wchar_t), &read, 0);
    CloseHandle(file);

    _p = _data;

    wchar_t line   [100];
    wchar_t section[100];
    wchar_t view   [100];
    wchar_t edit   [100];

    section[0] = view[0] = edit[0] = 0;

    while(get_line(line, std::size(line)))
    {
        wchar_t* p = skip_white_spaces(line);
        if(*p == 0 || *p == ';' || *p == '#') continue;

        if(*p == '[')
        {
            // section header
            ++p;
            wchar_t tmp[100];
            wchar_t* to = tmp;
            while(*p && *p != ']') *to++ = *p++;
            if(*p == ']')
            {
                *to = 0;
                if(tmp[0])//  
                {
                    if(section[0] && (edit[0] || view[0]))//  ,  
                    {
                        // association completely defined
                        Association& cur = _associations[_no_assosiations];
                        lstrcpyW(cur.format, section);//  
                        if(view[0]) lstrcpyW(cur.view, view);
                        if(edit[0]) lstrcpyW(cur.edit, edit);

                        section[0] = view[0] = edit[0] = 0;
                        _no_assosiations++;
                    }
                    lstrcpyW(section, tmp);
                }
            }
            continue;
        }
        // minimal line is "edit=X %f"
        if(lstrlenW(p) < 9) continue;

        if(p[0] == 'e' && p[1] == 'd' && p[2] == 'i' && p[3] == 't')
        {
            p += 4;
            p = skip_white_spaces(p);
            if(*p != '=') continue;
            ++p;
            p = skip_white_spaces(p);
            if(section[0] && *p) lstrcpyW(edit, p);
            continue;
        }

        if(p[0] == 'v' && p[1] == 'i' && p[2] == 'e' && p[3] == 'w')
        {
            p += 4;
            p = skip_white_spaces(p);
            if(*p != '=') continue;
            ++p;
            p = skip_white_spaces(p);
            if(section[0] && *p) lstrcpyW(view, p);
            continue;
        }
    }

    // check last section
    if(section[0] && (edit[0] || view[0]))
    {
        // association completely defined
        Association& cur = _associations[_no_assosiations];
        lstrcpyW(cur.format, section);
        if(view[0]) lstrcpyW(cur.view, view);
        if(edit[0]) lstrcpyW(cur.edit, edit);

        _no_assosiations++;
    }
    delete []_data;
    return true;
}
