#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>

#include "plugin.hpp"

#include "types.hpp"
#include "types_zxzip.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_zxzip.hpp"
#include "extract.hpp"
#include "lang.hpp"

#include "debug.hpp"

extern PluginStartupInfo startupInfo;
extern Options op;
extern const GUID MainGuid;

void addEndSlash(std::wstring &path);

bool Manager::extractFile(int fNum, BYTE* buffer, int size)
{
  int fSize = fileSize(files[fNum]);
  
  int packedSize = files[fNum].packedSize;
  BYTE* buf = new BYTE[256*files[fNum].noSecs];
  BYTE* pBuf = new BYTE[packedSize];

  if(!setPointer(pcFiles[fNum].offset)) goto error;
  if(!read(pBuf, packedSize)) goto error;
  
  switch(files[fNum].pMethod)
  {
    case 0:
      CopyMemory(buf, pBuf, fSize);
      break;

    case 1:
      unlz(pBuf, packedSize, buf, fSize);
      break;

    case 2:
      unshrink(pBuf, packedSize, buf, fSize);
      break;

    case 3:
      explode(pBuf, packedSize, buf, fSize, fSize < 0x1600, files[fNum].flags & 1);
      break;
  }

  if(crc32(buf, fSize) == files[fNum].crc)
  {
    CopyMemory(buffer, buf, size==0?fSize:size);
    delete[] buf;
    delete[] pBuf;
    return true;
  }

error:
  delete[] buf;
  delete[] pBuf;
  return false;
}

int Manager::getFiles(PluginPanelItem *panelItem,
                      int noItems,
                      int move,
                      const wchar_t *destPath,
                      int opMode)
{
  if(noItems == 0) return 1;

  std::wstring DestPathCopy(destPath);

  addEndSlash(DestPathCopy.data());

  wchar_t msg[100];
  if(noItems == 1)
    wsprintfW(msg, GetMsg(MExtract1), panelItem[0].FileName);
  else
    wsprintfW(msg, GetMsg(MExtract2), noItems);

  FarDialogItem dialogItems[]=
  {
    { DI_DOUBLEBOX,   3,  1, 72, 11, { FALSE },             nullptr, nullptr,                                   0, GetMsg(MExtract) },
    { DI_TEXT,        5,  2,  0,  0, { FALSE },             nullptr, nullptr,                                   0, msg },
    { DI_EDIT,        5,  3, 70,  3,  { TRUE }, L"xZXZIP_copy_path", nullptr,                         DIF_HISTORY, DestPathCopy.c_str() },
    { DI_TEXT,        3,  4,  0,  0, { FALSE },             nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_RADIOBUTTON, 5,  5,  0,  0, { FALSE },             nullptr, nullptr,                           DIF_GROUP, GetMsg(MHoBeta) },
    { DI_RADIOBUTTON, 5,  6,  0,  0, { FALSE },             nullptr, nullptr,                                   0, GetMsg(MSCL) },
    { DI_TEXT,        3,  7,  0,  0, { FALSE },             nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_CHECKBOX,    5,  8,  0,  0, { FALSE },             nullptr, nullptr,                                   0, GetMsg(MSkipHeaders) },
    { DI_TEXT,        3,  9,  0,  0, { FALSE },             nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_BUTTON,      0, 10,  0,  0, { FALSE },             nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MExtractButton) },
    { DI_BUTTON,      0, 10,  0,  0, { FALSE },             nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };
  
  switch(op.defaultFormat)
  {
    case 1:
              dialogItems[4].Selected = TRUE;
              break;
    case 2:
              dialogItems[5].Selected = TRUE;
              break;
    default:
              if(noItems == 1)
                dialogItems[4].Selected = TRUE;
              else
                dialogItems[5].Selected = TRUE;
  }
  
  bool singleFile = false, skipHeader = false;
  //    ,  
  if((opMode & OPM_SILENT) == 0)
  {
    std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                     &__uuidof(struct __declspec(uuid("{FD656077-E20C-4591-97D4-DB6638BD4029}")) TDlgGuid),
                                     -1, -1, 76, 13,
                                     NULL,
                                     dialogItems,
                                     std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
    if(Dlg.get() == INVALID_HANDLE_VALUE)
    {
        return -1;
    }
    intptr_t askCode = startupInfo.DialogRun(Dlg.get());
    if(askCode != 9) return -1;

    auto Sz = startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, nullptr);
    DestPathCopy.resize(Sz);

    FarDialogItemData ItemData = { sizeof(FarDialogItemData), DestPathCopy.size(), DestPathCopy.data() };
    startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);

    singleFile  = !startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0);
    skipHeader = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0);
  }
  
  //   ,   
  if(GetFileAttributesW(DestPathCopy.c_str()) == 0xFFFFFFFF)
  {
      for(wchar_t *c = DestPathCopy.data(); *c; c++)
      {
          if(*c != L' ')
          {
              for(; *c; c++)
              {
                  if(*c == L'\\')
                  {
                      *c = 0;
                      CreateDirectoryW(DestPathCopy.c_str(), NULL);
                      *c = L'\\';
                  }
              }
              CreateDirectoryW(DestPathCopy.c_str(), NULL);
              break;
          }
      }
  }
  if(!DestPathCopy.empty() && DestPathCopy.back() != ':') addEndSlash(DestPathCopy);

  DWORD noBytesWritten;
  DWORD checkSum     = 0;
  
  int buttonStatus = 0; //   overwrite_all/skip_all
  
  HANDLE file;
  wchar_t   name[300];
  
  if(singleFile)
  {
    wsprintfW(name, L"%s%s", DestPathCopy.c_str(), panelItem[0].FileName);

    wchar_t* ptr = name + lstrlenW(name);
    while(*ptr != L'.') { --ptr; }
    *ptr = 0;

    if(skipHeader)
      lstrcatW(name, L".bin");
    else
      lstrcatW(name, L".scl");
    int r = createFile(file, name, buttonStatus);
    if(r == 0 || r == -1) return -1;
    if(!skipHeader) checkSum = writeSCLHeader(file, noItems);
  }
  
  int exitCode = 1;
  
  if(!openHostFile()) return 0;
  
  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);

  DWORD offset = noItems*sizeof(SCLHdr) + 8 + 1;
  
  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    //   
    int fNum;
    for(fNum = 0; fNum < noFiles; ++fNum)
      if(!lstrcmpW(panelItem[iNum].FileName, pcFiles[fNum].name)) break;
    
    //     
    if(noItems == 1 && ((opMode & OPM_VIEW) || (opMode & OPM_EDIT)) && pcFiles[fNum].skipHeader) skipHeader = true;
    
    if(!singleFile)
    {
      wsprintfW(name, L"%s%s", DestPathCopy.c_str(), panelItem[iNum].FileName);
      if(skipHeader) lstrcatW(name, L".bin");
      int r = createFile(file, name, buttonStatus);
      if(r == -1)
      {
        exitCode = -1;
        break;
      }
      if(r == 0)
      {
        exitCode = 0;
        continue;
      }
      if(!skipHeader)
      {
        HoHdr hdr;
        CopyMemory(hdr.name, files[fNum].name, 8);
        
        hdr.type     = files[fNum].type;
        hdr.start    = files[fNum].start;
        hdr.size     = files[fNum].size;
        hdr.reserved = 0;
        hdr.noSecs   = files[fNum].noSecs;
        hdr.checkSum = calculateCheckSum(hdr);
        WriteFile(file, &hdr, sizeof(HoHdr), &noBytesWritten, NULL);
      }
    }
    else
    {
      if(!skipHeader)
      {
        SCLHdr hdr;
        CopyMemory(hdr.name, files[fNum].name, 8);

        hdr.type   = files[fNum].type;
        hdr.start  = files[fNum].start;
        hdr.size   = files[fNum].size;
        hdr.noSecs = files[fNum].noSecs;

        //  
        SetFilePointer(file, iNum*sizeof(SCLHdr)+8+1, NULL, FILE_BEGIN);
        WriteFile     (file, &hdr, sizeof(SCLHdr), &noBytesWritten, NULL);
        checkSum += calculateCheckSum((BYTE*)&files[fNum], sizeof(SCLHdr));
        
        SetFilePointer(file, offset, NULL, FILE_BEGIN);
        offset += 256*files[fNum].noSecs;
      }
    }
    
    const wchar_t* msgItems[2];
    msgItems[0] = GetMsg(MExtracting);
    wchar_t  shortName[51] = L"...";
    msgItems[1] = name;
    if(lstrlenW(name) > 47)
    {
      lstrcatW(shortName, name + lstrlenW(name) - 47);
      msgItems[1] = shortName;
    }

    if((opMode & OPM_SILENT) == 0)
      messageBox(0, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 0);

    //  
    BYTE* buf = new BYTE[256*files[fNum].noSecs];

    if(!extractFile(fNum, buf, 0))
    {
      const wchar_t*         msgItems[3];
      msgItems[0] = GetMsg(MCRCError);
      msgItems[1] = panelItem[iNum].FileName;
      msgItems[2] = GetMsg(MOk);
      
      messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    }
    else
      //     
      panelItem[iNum].Flags ^= PPIF_SELECTED;
    
    checkSum += calculateCheckSum(buf, 256*files[fNum].noSecs);
    
    DWORD noBytesWritten;
    if(skipHeader)
      WriteFile(file, buf, fileSize(files[fNum]), &noBytesWritten, NULL);
    else
      WriteFile(file, buf, 256*files[fNum].noSecs, &noBytesWritten, NULL);

    delete[] buf;

    if(!singleFile) CloseHandle(file);
    
  }
  closeHostFile();

  startupInfo.RestoreScreen(screen);

  if(singleFile)
  {
    if(!skipHeader)
      WriteFile(file, &checkSum, sizeof(checkSum), &noBytesWritten, NULL);
    CloseHandle(file);
  }
  return exitCode;
}
